/*
 * SnapShotReader.java
 * Created on 2011/10/13
 *
 * Copyright (C) 2011-2013 Nippon Telegraph and Telephone Corporation
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

package jp.co.ntt.oss.heapstats.gui.snapshot;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;

import jp.co.ntt.oss.heapstats.container.ObjectData;
import jp.co.ntt.oss.heapstats.container.SnapShot;
import jp.co.ntt.oss.heapstats.container.SnapShotHeader;
import jp.co.ntt.oss.heapstats.gui.GuiCommon;

/**
 * This class will load the temporary files java heap information.
 */
public class SnapShotReader {
    /**
     * It is the column number that indicates the date snapshot of the header
     * file.
     */
    private static final int COL_HEADER_DATE = 0;

    /**
     * It is the column number that indicates the number of objects in the
     * header file.
     */
    private static final int COL_HEADER_NUMENTRY = 1;

    /**
     * It is the column number that indicates the reason for taking a snapshot.
     */
    private static final int COL_HEADER_CAUSE = 2;

    /**
     * It is a number that indicates the reason for generation GC column header.
     */
    private static final int COL_HEADER_GCCAUSE = 3;

    /**
     * It is the column number that indicates the number of occurrences Full-GC.
     */
    private static final int COL_HEADER_FULL = 4;

    /**
     * It is the column number that indicates the number of occurrences
     * Young-GC.
     */
    private static final int COL_HEADER_YOUNG = 5;

    /** It is the column number that indicates the processing time GC. */
    private static final int COL_HEADER_GCTIME = 6;

    /** It is the column number that indicates the New heap size. */
    private static final int COL_HEADER_NEW = 7;

    /** It is the column number that indicates the Old heap size. */
    private static final int COL_HEADER_OLD = 8;

    /** It is the column number that indicates the total size of the heap. */
    private static final int COL_HEADER_TOTAL = 9;

    /** It is the column number that indicates the metaspace usage. */
    private static final int COL_HEADER_METASPACE_USAGE = 10;

    /** It is the column number that indicates the metaspace capacity. */
    private static final int COL_HEADER_METASPACE_CAPACITY = 11;

    /**
     * Is a constant for the column number of the class tag of temporary files
     * class information.
     */
    private static final int COL_CLASS_CLASSTAG = 0;

    /**
     * Is a constant for the column number for the class name of the temporary
     * file information class.
     */
    private static final int COL_CLASS_NAME = 1;

    /**
     * Is a constant for the column number of the class loader of the temporary
     * file information.
     */
    private static final int COL_CLASS_LOADER = 2;

    /**
     * The column number of the class loader is a constant for temporary files
     * tag information.
     */
    private static final int COL_CLASS_LOADERTAG = 3;

    /**
     * It is a constant of the column number of the number of instances of the
     * temporary file information.
     */
    private static final int COL_CLASS_COUNT = 4;

    /**
     * It is a constant of the column number of the heap usage of temporary file
     * information.
     */
    private static final int COL_CLASS_SIZE = 5;

    /**
     * The column number of the class is a constant for a child tag of the
     * temporary file information.
     */
    private static final int COL_CHILD_CLASSTAG = 2;

    /**
     * It is a constant of the column number of the number of instances of the
     * temporary file information.
     */
    private static final int COL_CHILD_COUNT = 3;

    /**
     * It is a constant of the column number of the heap usage of temporary file
     * information.
     */
    private static final int COL_CHILD_SIZE = 4;

    /**
     * It is a constant of the column number of the name of the class loader of
     * the temporary file information.
     */
    private static final int COL_CHILD_LOADERNAME = 5;

    /** Hold the absolute path of the parent directory of the snapshot file. */
    private String dirName;

    /**
     * Create a SnapShotReader.
     *
     * @param fname the temporary file path java heap information.
     */
    public SnapShotReader(final String fname) {
        dirName = fname;
    }

    /**
     * Gets the header portion of the snapshot file.
     *
     * @return Returns the header portion of the snapshot file.
     * @throws IOException If an I/O error occurs
     */
    public final SnapShotHeader getHeader() throws IOException {
        BufferedReader read = null;
        String[] line = null;
        SnapShotHeader header = new SnapShotHeader();
        try {
            read = new BufferedReader(new FileReader(dirName + File.separator
                    + GuiCommon.SNAPSHOT_TEMP_NAME_HEADER));
            String tmp = read.readLine();
            if (tmp != null) {
                line = tmp.split(",");
            }
        } finally {
            if (read != null) {
                read.close();
            }
        }

        if (line != null && line.length == 12) {
            header.setSnapShotDate(new Date(Long
                    .parseLong(line[COL_HEADER_DATE])));
            header.setNumEntries(Long.parseLong(line[COL_HEADER_NUMENTRY]));
            header.setCause(Integer.parseInt(line[COL_HEADER_CAUSE]));
            header.setGcCause(line[COL_HEADER_GCCAUSE]);
            header.setFullCount(Long.parseLong(line[COL_HEADER_FULL]));
            header.setYngCount(Long.parseLong(line[COL_HEADER_YOUNG]));
            header.setGcTime(Long.parseLong(line[COL_HEADER_GCTIME]));
            header.setNewHeap(Long.parseLong(line[COL_HEADER_NEW]));
            header.setOldHeap(Long.parseLong(line[COL_HEADER_OLD]));
            header.setTotalSize(Long.parseLong(line[COL_HEADER_TOTAL]));
            header.setMetaspaceUsage(Long.parseLong(line[COL_HEADER_METASPACE_USAGE]));
            header.setMetaspaceCapacity(Long.parseLong(line[COL_HEADER_METASPACE_CAPACITY]));
        }

        return header;
    }

    /**
     * Return the snapshot header and objects data.
     *
     * @return Return the SnapShot information.
     * @throws IOException If an I/O error has occurred.
     * @throws ClassNotFoundException Class of a serialized object cannot be
     *         found.
     */
    public final synchronized SnapShot getSnapShot() throws IOException,
            ClassNotFoundException {
        SnapShot snapShot = getClassEntry();
        snapShot.setHeader(getHeader());

        return snapShot;
    }

    /**
     * Gets the information to the list of objects.
     *
     * @return Return the SnapShot information.
     * @throws IOException If an I/O error has occurred.
     */
    public final synchronized SnapShot getClassEntry() throws IOException {
        BufferedReader read = null;
        long heapUsage = 0;
        SnapShot snapShot = new SnapShot();
        snapShot.setElements(new HashMap<Long, ObjectData>());

        try {
            read = new BufferedReader(new FileReader(dirName + File.separator
                    + GuiCommon.SNAPSHOT_TEMP_NAME_CLASS));
            String line = read.readLine();

            while (line != null) {
                ObjectData data = setClassData(line);
                if (data != null) {
                    snapShot.getElements().put(data.getTag(), data);
                    heapUsage += data.getTotalSize();
                }

                line = read.readLine();
            }
        } finally {
            if (read != null) {
                read.close();
            }
        }

        snapShot.setHeapUsage(heapUsage);

        return snapShot;
    }

    /**
     * Get information of the specified object.
     *
     * @param tag Specifies the ID that is unique to the object that has been
     *        granted.
     * @return Return the object information.
     * @throws IOException If an I/O error has occurred.
     */
    public final synchronized ObjectData getClassEntry(final long tag)
            throws IOException {
        BufferedReader read = null;
        ObjectData data = null;

        try {
            read = new BufferedReader(new FileReader(dirName + File.separator
                    + GuiCommon.SNAPSHOT_TEMP_NAME_CLASS));
            String line = read.readLine();

            while (line != null) {
                ObjectData tmpData = setClassData(line);
                if (tmpData != null && tmpData.getTag() == tag) {
                    data = tmpData;
                    break;
                }

                line = read.readLine();
            }
        } finally {
            if (read != null) {
                read.close();
            }
        }

        return data;
    }

    /**
     * Gets the information to the list of objects.
     *
     * @param tagList Specifies a list of ID that is unique to the object that
     *        has been granted.
     * @return Return the SnapShot information.
     * @throws IOException If an I/O error has occurred.
     */
    public final synchronized SnapShot getClassEntry(final List<Long> tagList)
            throws IOException {
        BufferedReader read = null;
        SnapShot snapShot = new SnapShot();
        snapShot.setElements(new HashMap<Long, ObjectData>());

        try {
            read = new BufferedReader(new FileReader(dirName + File.separator
                    + GuiCommon.SNAPSHOT_TEMP_NAME_CLASS));
            String line = read.readLine();

            while (line != null) {

                ObjectData data = setClassData(line);
                if (checkClassTag(data, tagList)) {
                    snapShot.getElements().put(data.getTag(), data);
                }

                line = read.readLine();
            }
        } finally {
            if (read != null) {
                read.close();
            }
        }

        return snapShot;
    }

    /**
     * Gets information about the object that the specified object is holding.
     *
     * @param tag Specifies the ID that is unique to the object that has been
     *        granted.
     * @return Return the SnapShot information.
     * @throws IOException If an I/O error has occurred.
     */
    public final synchronized SnapShot getToChildren(final long tag)
            throws IOException {
        return getReferenceClass(tag, true);
    }

    /**
     * Gets a list object that refers to the given object.
     *
     * @param tag Specifies the ID that is unique to the object that has been
     *        granted.
     * @return Return the SnapShot information.
     * @throws IOException If an I/O error has occurred.
     */
    public final synchronized SnapShot getToParents(final long tag)
            throws IOException {
        return getReferenceClass(tag, false);
    }

    /**
     * Gets the information in the list of objects that are related to the
     * specified object.
     *
     * @param tag Specifies the ID that is unique to the object that has been
     *        granted.
     * @param child If the child class to set the true
     * @return Return the SnapShot information.
     * @throws IOException If an I/O error has occurred.
     */
    private SnapShot getReferenceClass(final long tag, final boolean child)
            throws IOException {
        BufferedReader read = null;
        SnapShot snapShot = new SnapShot();

        List<ObjectData> objectList = new ArrayList<ObjectData>();
        List<Long> tagList = new ArrayList<Long>();

        try {
            StringBuilder path = new StringBuilder();
            path.append(dirName);
            path.append(File.separator);
            path.append(GuiCommon.SNAPSHOT_TEMP_NAME_REFERENCE);

            read = new BufferedReader(new FileReader(path.toString()));

            if (child) {
                snapShot.setToChildren(new HashMap<Long, List<ObjectData>>());
            } else {
                snapShot.setToParents(new HashMap<Long, List<ObjectData>>());
            }

            String line = read.readLine();

            while (line != null) {
                String[] dataList = line.split(",");
                if (Long.parseLong(dataList[0]) == tag) {
                    ObjectData data = new ObjectData();

                    data.setTag(Long.parseLong(dataList[COL_CHILD_CLASSTAG]
                            .trim()));
                    data.setCount(Long.parseLong(dataList[COL_CHILD_COUNT]
                            .trim()));
                    data.setTotalSize(Long.parseLong(dataList[COL_CHILD_SIZE]
                            .trim()));
                    data.setLoaderName(dataList[COL_CHILD_LOADERNAME].trim());

                    if ((child && dataList[1]
                            .equals(GuiCommon.SNAPSHOT_TEMP_CHILD))
                            || (!child && dataList[1]
                                    .equals(GuiCommon.SNAPSHOT_TEMP_PARENT))) {
                        objectList.add(data);
                        tagList.add(data.getTag());
                    }
                }

                line = read.readLine();
            }
        } finally {
            if (read != null) {
                read.close();
            }
        }

        // Gets the class name
        if (tagList.size() > 0) {
            SnapShot tmp = getClassEntry(tagList);
            for (ObjectData entry : objectList) {
                ObjectData data = tmp.getElements().get(entry.getTag());
                if (data != null) {
                    entry.setName(data.getName());
                } else {
                    entry.setName("-");
                }
            }

            if (child) {
                snapShot.getToChildren().put(tag, objectList);
            } else {
                snapShot.getToParents().put(tag, objectList);
            }
        }

        return snapShot;
    }

    /**
     * Set the object provides information about the objects that are separated
     * by commas obtained from the temporary file.
     *
     * @param dataLine Information objects that are separated by commas obtained
     *        from the temporary file.
     * @return the object information.
     */
    private ObjectData setClassData(final String dataLine) {
        ObjectData data = null;
        String[] dataList = dataLine.split(",");

        if (dataList.length == 6) {
            data = new ObjectData();
            data.setTag(Long.valueOf(dataList[COL_CLASS_CLASSTAG]));
            data.setName(dataList[COL_CLASS_NAME].trim());
            data.setClassLoader(Long.parseLong(dataList[COL_CLASS_LOADER]
                    .trim()));
            data.setClassLoaderTag(Long.parseLong(dataList[COL_CLASS_LOADERTAG]
                    .trim()));
            data.setCount(Long.parseLong(dataList[COL_CLASS_COUNT].trim()));
            data.setTotalSize(Long.parseLong(dataList[COL_CLASS_SIZE].trim()));
        }

        return data;
    }

    /**
     * Checks whether the tag is present in the list of the specified object.
     *
     * @param data Class that holds the information of the object being loaded
     *        to the heap.
     * @param tagList Specifies a list of ID that is unique to the object that
     *        has been granted.
     * @return Return the check result. true:Exist in the tag list.
     */
    private boolean checkClassTag(final ObjectData data,
            final List<Long> tagList) {
        if (data == null) {
            return false;
        }

        for (long tag : tagList) {
            if (data.getTag() == tag) {
                return true;
            }
        }

        return false;
    }
}
